import os
import sys
import subprocess as sp

from genomepy.plugin import Plugin
from genomepy.utils import mkdir_p, cmd_ok, run_index_cmd

class Minimap2Plugin(Plugin):
    def after_genome_download(self, genome):
        if not cmd_ok("minimap2"):
            return

        # Create index dir
        index_dir = genome.props["minimap2"]["index_dir"]
        index_name =  genome.props["minimap2"]["index_name"] 
        mkdir_p(index_dir)

        # Create index
        cmd = "minimap2 -d {} {}".format(index_name, genome.filename)
        run_index_cmd("minimap2", cmd)
                
    def get_properties(self, genome):
        props = {
            "index_dir": os.path.join(
                os.path.dirname(genome.filename), "index", "minimap2"
                ),
            "index_name": os.path.join(
                os.path.dirname(genome.filename), 
                "index", "minimap2", "{}.mmi".format(genome.name)
                ),
            }
        return props



"""Search, download and use genome FASTA files."""
from genomepy.functions import (
        list_available_providers, list_available_genomes,
        list_installed_genomes, search, install_genome,
        Genome)
from genomepy.__about__ import __version__, __author__



import re
import sys
from genomepy.plugin import Plugin
try:
    from urllib.request import urlopen
except:
    from urllib import urlopen

class BlacklistPlugin(Plugin):

    http_dict = {
        "ce10" : "http://mitra.stanford.edu/kundaje/akundaje/release/blacklists/ce10-C.elegans/ce10-blacklist.bed.gz",
        "dm3"  : "http://mitra.stanford.edu/kundaje/akundaje/release/blacklists/dm3-D.melanogaster/dm3-blacklist.bed.gz",
        "hg38" : "http://mitra.stanford.edu/kundaje/akundaje/release/blacklists/hg38-human/hg38.blacklist.bed.gz",
        "mm9"  : "http://mitra.stanford.edu/kundaje/akundaje/release/blacklists/mm9-mouse/mm9-blacklist.bed.gz",
        "mm10" : "http://mitra.stanford.edu/kundaje/akundaje/release/blacklists/mm10-mouse/mm10.blacklist.bed.gz",
    }

    def after_genome_download(self, genome):
        props = self.get_properties(genome)
        fname = props["blacklist"]
       
        link = self.http_dict.get(genome.name)
        if link is None: 
            sys.stderr.write("No blacklist found for {}\n".format(genome.name))
        try:
            sys.stderr.write("Downloading blacklist {}\n".format(link))
            response = urlopen(link)
            with open(fname, "wb") as bed:
                bed.write(response.read())
        except Exception as e:
            print(e)
            print("Could not download blacklist file from {}".format(link))
    
    def get_properties(self, genome):
        props = {
               "blacklist": genome.filename.replace(".fa", ".blacklist.bed.gz"),
               }
        return props



import os
import sys
import subprocess as sp

from genomepy.plugin import Plugin
from genomepy.utils import mkdir_p, cmd_ok, run_index_cmd

class GmapPlugin(Plugin):
    def after_genome_download(self, genome):
        if not cmd_ok("gmap_build"):
            return

        # Create index dir
        index_dir = genome.props["gmap"]["index_dir"]
        index_name =  genome.props["gmap"]["index_name"] 
        mkdir_p(index_dir)

        # Create index
        cmd = "gmap_build -D {} -d {} {}".format(
                index_dir, genome.name, genome.filename)
        run_index_cmd("gmap", cmd)
         
    def get_properties(self, genome):
        props = {
            "index_dir": os.path.join(
                os.path.dirname(genome.filename), "index" , "gmap"
                ),
            "index_name": os.path.join(
                os.path.dirname(genome.filename), "index", "gmap", genome.name
                ),
            }
        return props



import os
import sys
import subprocess as sp

from genomepy.plugin import Plugin
from genomepy.utils import mkdir_p, cmd_ok, run_index_cmd

class BwaPlugin(Plugin):
    def after_genome_download(self, genome):
        if not cmd_ok("bwa"):
            return
        
        # Create index dir
        index_dir = genome.props["bwa"]["index_dir"]
        index_fa =  genome.props["bwa"]["index_name"] 
        mkdir_p(index_dir)

        if not os.path.exists(index_fa):
            os.symlink(genome.filename, index_fa)

        cmd = "bwa index {}".format(index_fa)
        run_index_cmd("bwa", cmd)
               
    def get_properties(self, genome):
        props = {
            "index_dir": os.path.join(
                os.path.dirname(genome.filename), "index", "bwa"
                ),
            "index_name": os.path.join(
                os.path.dirname(genome.filename), 
                "index","bwa", "{}.fa".format(genome.name)
                ),
            }
        return props



"""Utility functions."""
import errno
import os
import re
import sys
import subprocess as sp

from pyfaidx import Fasta

def generate_gap_bed(fname, outname):
    """ Generate a BED file with gap locations.

    Parameters
    ----------
    fname : str
        Filename of input FASTA file.

    outname : str
        Filename of output BED file.
    """ 
    f = Fasta(fname)
    with open(outname, "w") as bed:
        for chrom in f.keys():
            for m in re.finditer(r'N+', f[chrom][:].seq):
                bed.write("{}\t{}\t{}\n".format(chrom, m.start(0), m.end(0)))

def generate_sizes(name, genome_dir):
    """Generate a sizes file with length of sequences in FASTA file."""
    fa = os.path.join(genome_dir, name, "{}.fa".format(name))
    sizes = fa + ".sizes"
    g = Fasta(fa)
    with open(sizes, "w") as f:
        for seqname in g.keys():
            f.write("{}\t{}\n".format(seqname, len(g[seqname])))

def filter_fasta(infa, outfa, regex=".*", v=False, force=False):
    """Filter fasta file based on regex.

    Parameters
    ----------
    infa : str
        Filename of input fasta file.
    
    outfa : str
        Filename of output fasta file. Cannot be the same as infa.

    regex : str, optional
        Regular expression used for selecting sequences.

    v : bool, optional
        If set to True, select all sequence *not* matching regex.

    force : bool, optional
        If set to True, overwrite outfa if it already exists.

    Returns
    -------
        fasta : Fasta instance
            pyfaidx Fasta instance of newly created file
    """
    if infa == outfa:
        raise ValueError("Input and output FASTA are the same file.")

    if os.path.exists(outfa):
        if force:
            os.unlink(outfa)
            if os.path.exists(outfa + ".fai"):
                os.unlink(outfa + ".fai")
        else:
            raise ValueError(
                    "{} already exists, set force to True to overwrite".format(outfa))
            
    filt_function = re.compile(regex).search
    fa = Fasta(infa, filt_function=filt_function)
    seqs = fa.keys()
    if v:
        original_fa = Fasta(infa)
        seqs = [s for s in original_fa.keys() if s not in seqs]
        fa = original_fa
    
    if len(seqs) == 0:
        raise ValueError("No sequences left after filtering!")

    with open(outfa, "w") as out:
        for chrom in seqs:
            out.write(">{}\n".format(fa[chrom].name))
            out.write("{}\n".format(fa[chrom][:].seq))

    return Fasta(outfa)

def mkdir_p(path):
    """ 'mkdir -p' in Python """
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

def cmd_ok(cmd):
    """Returns True if cmd can be run.
    """ 
    try:
        sp.check_call(cmd, stderr=sp.PIPE, stdout=sp.PIPE)
    except sp.CalledProcessError:
        # bwa gives return code of 1 with no argument
        pass
    except:
        sys.stderr.write("{} not found, skipping\n".format(cmd))
        return False
    return True

def run_index_cmd(name, cmd):
    """Run command, show errors if the returncode is non-zero."""
    sys.stderr.write("Creating {} index...\n".format(name))
    # Create index
    p = sp.Popen(cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()
    if p.returncode != 0:
        sys.stderr.write("Index for {} failed\n".format(name))
        sys.stderr.write(stdout)
        sys.stderr.write(stderr)

def get_localname(name, localname):
    """Returns localname if localname is not None, else returns name."""
    if localname is None:
        return name.replace(' ', '_')
    else:
        return localname.replace(" ", "_")



import os
import sys
import re

import norns

config = norns.config("genomepy", default="cfg/default.yaml")

class Plugin(object):
    """Plugin base class.
    """
    active = False

    def name(self):
        n = type(self).__name__.replace("Plugin", "")
        return convert(n)

    def activate(self):
        self.active = True

    def deactivate(self):
        self.active = False

    def after_genome_download(self, genome):
        raise NotImplementedError("plugin should implement this method")

    def get_properties(self, genome):
        raise NotImplementedError("plugin should implement this method")

def find_plugins():
    """Locate and initialize all available plugins.
    """ 
    plugin_dir = os.path.dirname(os.path.realpath(__file__))
    plugin_dir = os.path.join(plugin_dir, "plugins")
    plugin_files = [x[:-3] for x in os.listdir(plugin_dir) if x.endswith(".py")]
    for plugin in plugin_files:
        __import__("genomepy.plugins.{}".format(plugin))

def convert(name):
    """Convert CamelCase to underscore

    Parameters
    ----------
    name : str
        Camelcase string

    Returns
    -------
    name : str
        Converted name
    """ 
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

def init_plugins():
    """Return dictionary of available plugins

    Returns
    -------
    plugins : dictionary
        key is plugin name, value Plugin object
    """ 
    find_plugins()
    d = {}
    for c in Plugin.__subclasses__():
        ins = c()
    
        if ins.name() in config.get("plugin", []):
            ins.activate()
        
        d[ins.name()] = ins
    
    return d

def activate(name):
    """Activate plugin.

    Parameters
    ----------
    name : str
        Plugin name.
    """
    if name in plugins:
        plugins[name].activate()
    else:
        raise Exception("plugin {} not found".format(name))

def deactivate(name):
    """Deactivate plugin.

    Parameters
    ----------
    name : str
        Plugin name.
    """
    if name in plugins:
        plugins[name].deactivate()
    else:
        raise Exception("plugin {} not found".format(name))

def get_active_plugins():
    """Returns all active plugin instances.
    """ 
    return [inst for name, inst in plugins.items() if inst.active]

plugins = init_plugins()



"""Module-level functions."""
import bisect
import os
import glob
import random
import re
import shutil
from tempfile import mkdtemp

from appdirs import user_config_dir
from pyfaidx import Fasta,Sequence
from genomepy.provider import ProviderBase
from genomepy.plugin import get_active_plugins, init_plugins, activate
from genomepy.utils import get_localname
import norns

config = norns.config("genomepy", default="cfg/default.yaml")

# Python 2
try:
    FileNotFoundError
except NameError:
    # pylint: disable=redefined-builtin
    FileNotFoundError = IOError

def manage_config(cmd, *args):
    """Manage genomepy config file."""
    if cmd == "file":
        print(config.config_file)
    elif cmd == "show":
        with open(config.config_file) as f:
            print(f.read())
    elif cmd == "generate":
        fname = os.path.join(
                user_config_dir("genomepy"), "{}.yaml".format("genomepy")
            )
        
        if not os.path.exists(user_config_dir("genomepy")):
            os.makedirs(user_config_dir("genomepy"))
        
        with open(fname, "w") as fout:
            with open(config.config_file) as fin:
                fout.write(fin.read())
        print("Created config file {}".format(fname))

def list_available_genomes(provider=None):
    """
    List all available genomes.

    Parameters
    ----------
    provider : str, optional
        List genomes from specific provider. Genomes from all
        providers will be returned if not specified.

    Returns
    -------
    list with genome names
    """
    if provider:
        providers = [ProviderBase.create(provider)]
    else:
        # if provider is not specified search all providers
        providers = [ProviderBase.create(p) for 
                        p in ProviderBase.list_providers()]

    for p in providers:
        for row in p.list_available_genomes():
            yield [p.name] + list(row)

def list_available_providers():
    """
    List all available providers.

    Returns
    -------
    list with provider names
    """
    return ProviderBase.list_providers()

def _is_genome_dir(dirname):
    """
    Check if a directory contains a fasta file
    
    Parameters
    ----------
    dirname : str
        Directory name

    Returns
    ------
    bool 
    """
    return len(glob.glob("{}/*.fa".format(dirname))) > 0

def list_installed_genomes(genome_dir=None):
    """
    List all available genomes.

    Parameters
    ----------
    genome_dir : str
        Directory with installed genomes.

    Returns
    -------
    list with genome names
    """
    if not genome_dir:
        genome_dir = config.get("genome_dir", None)
    if not genome_dir:
        raise norns.exceptions.ConfigError("Please provide or configure a genome_dir")

    return [f for f in os.listdir(genome_dir) if 
            _is_genome_dir(genome_dir + "/" + f)]

def search(term, provider=None):
    """
    Search for a genome.

     If provider is specified, search only that specific provider, else 
     search all providers. Both the name and description are used for the 
     search. Seacrch term is case-insensitive.

    Parameters
    ----------
    term : str
        Search term, case-insensitive.
    
    provider : str , optional
        Provider name

    Yields
    ------
    tuple
        genome information (name/identfier and description)
    """
    if provider:
        providers = [ProviderBase.create(provider)]
    else:
        # if provider is not specified search all providers
        providers = [ProviderBase.create(p) for 
                        p in ProviderBase.list_providers()]

    for p in providers:
        for row in p.search(term):
            yield [x.encode('latin-1') for x in [p.name] + list(row)]

def install_genome(name, provider, version=None, genome_dir=None, localname=None, mask="soft", regex=None, invert_match=False, annotation=False):
    """
    Install a genome.

    Parameters
    ----------
    name : str
        Genome name

    provider : str
        Provider name

    version : str
        Version (only for Ensembl)

    genome_dir : str , optional
        Where to store the fasta files
    
    localname : str , optional
        Custom name for this genome.

    mask : str , optional
        Default is 'soft', specify 'hard' for hard masking.

    regex : str , optional
        Regular expression to select specific chromosome / scaffold names.

    invert_match : bool , optional
        Set to True to select all chromosomes that don't match the regex.

    annotation : bool , optional
        If set to True, download gene annotation in BED and GTF format.
    """
    if not genome_dir:
        genome_dir = config.get("genome_dir", None)
    if not genome_dir:
        raise norns.exceptions.ConfigError("Please provide or configure a genome_dir")
   
    genome_dir = os.path.expanduser(genome_dir)
    localname = get_localname(name, localname)
    
    # Download genome from provider
    p = ProviderBase.create(provider)
    p.download_genome(
        name, 
        genome_dir, 
        version=version,
        mask=mask, 
        localname=localname, 
        regex=regex, 
        invert_match=invert_match)

    if annotation:
        # Download annotation from provider
        p.download_annotation(name, genome_dir, localname=localname, version=version)

    g = Genome(localname, genome_dir=genome_dir)
    for plugin in get_active_plugins():
        plugin.after_genome_download(g)

    generate_env()

def get_track_type(track):
    region_p = re.compile(r'^(.+):(\d+)-(\d+)$')
    if type(track) == type([]):
        if region_p.search(track[0]):
            return "interval"

    with open(track) as fin:
        line = fin.readline().strip()
    if region_p.search(line):
        return "interval"
    return "bed"

def _weighted_selection(l, n):
    """
        Selects  n random elements from a list of (weight, item) tuples.
        Based on code snippet by Nick Johnson
    """
    cuml = []
    items = []
    total_weight = 0.0
    for weight, item in l:
        total_weight += weight
        cuml.append(total_weight)
        items.append(item)

    return [items[bisect.bisect(cuml, random.random()*total_weight)] for _ in range(n)]

def generate_exports():
	"""Print export commands for setting environment variables.
	"""
	env = []
	for name in list_installed_genomes():
	    try:
	        g = Genome(name)
	        env_name = re.sub(r'[^\w]+', "_", name).upper()
	        env.append("export {}={}".format(env_name, g.filename))
	    except:
	        pass
	return env

def generate_env(fname=None):
    """Generate file with exports. 

    By default this is in .config/genomepy/exports.txt.

    Parameters
    ----------
    fname: strs, optional
        Name of the output file.
    """
    config_dir = user_config_dir("genomepy")
    if os.path.exists(config_dir):
        fname = os.path.join(config_dir, "exports.txt")
        with open(fname, "w") as fout:
            for env in generate_exports():
                fout.write("{}\n".format(env))

class Genome(Fasta):
    """
    Get pyfaidx Fasta object of genome

    Parameters
    ----------
    name : str
        Genome name
    
    genome_dir : str
        Genome installation directory

    Returns
    -------
    pyfaidx.Fasta object
    """
    def __init__(self, name, genome_dir=None):
        
        try:
            super(Genome, self).__init__(name)
            self.name = os.path.basename(name)
        except:
            if os.path.isdir(name) and len(glob.glob(os.path.join(name, "*.fa"))) == 1 and genome_dir is not None:
                fname = glob.glob(os.path.join(name, "*.fa"))[0]
                name = os.path.basename(fname)  
            else:
                if not genome_dir:
                    genome_dir = config.get("genome_dir", None)
                if not genome_dir:
                    raise norns.exceptions.ConfigError("Please provide or configure a genome_dir")
        
                genome_dir = os.path.expanduser(genome_dir)
                if not os.path.exists(genome_dir):
                    raise FileNotFoundError(
                            "genome_dir {} does not exist".format(genome_dir)
                            )
    
                pattern = os.path.join(genome_dir, name, "*.fa")
                fnames = glob.glob(pattern)
                if len(fnames) == 0:
                    raise FileNotFoundError(
                            "no *.fa files found in genome_dir {}".format(
                                os.path.join(genome_dir, name)
                                )
                            )
                elif len(fnames) > 1:
                    fname = os.path.join(genome_dir, name, "{}.fa".format(name))
                    if fname not in fnames:
                        raise Exception("More than one FASTA file found, no {}.fa!".format(name))
                else:
                    fname = fnames[0]
    
            super(Genome, self).__init__(fname)
        
            self.name = name
        
        self._gap_sizes = None
        self.props = {}

        for plugin in get_active_plugins():
            self.props[plugin.name()] = plugin.get_properties(self) 

    def _bed_to_seqs(self, track, stranded=False, extend_up=0, extend_down=0):
        BUFSIZE = 10000
        with open(track) as fin:
            lines = fin.readlines(BUFSIZE)
            while lines:
                for line in lines:
                    if line.startswith("#") or line.startswith("track"):
                        continue
    
                    vals = line.strip().split("\t")
                    try:
                        start, end = int(vals[1]), int(vals[2])
                    except ValueError:
                        raise
    
                    rc = False
                    if stranded:
                        try:
                            rc = vals[5] == "-"
                        except IndexError:
                            pass
    
                    starts = [start]
                    ends = [end]
                    
                    chrom = vals[0]
                    
                    # BED12
                    if len(vals) == 12:
                        starts = [int(x) for x in vals[11].split(",")[:-1]]
                        sizes = [int(x) for x in vals[10].split(",")[:-1]]
                        starts = [start + x  for x in starts]
                        ends = [start + size  for start,size in zip(starts, sizes)]
                    name = "{}:{}-{}".format(chrom, start, end)    
                    try:
                        name = " ".join((name, vals[3]))
                    except: 
                        pass

                    starts = [start + 1 for start in starts]
                    
                    # extend
                    if extend_up:
                        if rc:
                            ends[-1] += extend_up
                        else:
                            starts[0] -= extend_up
                    if extend_down:
                        if rc:
                            starts[0] -= extend_down
                        else:
                            ends[-1] += extend_down
    
                    intervals = zip(starts, ends)
                    seq = self.get_spliced_seq(chrom, intervals, rc)
                    yield Sequence(name, seq.seq)
                
                lines = fin.readlines(BUFSIZE)

    def _region_to_seqs(self, track, extend_up=0, extend_down=0):
        BUFSIZE = 10000
        if type(track) == type([]):
            for name in track:
                    chrom, coords = name.split(":")
                    start, end = [int(c) for c in coords.split("-")]
                    start += 1
                    start -= extend_up
                    end += extend_down
                    seq = self.get_seq(chrom, start, end)
                    yield Sequence(name, seq.seq)
        else:
            with open(track) as fin:
                lines = fin.readlines(BUFSIZE)
                while lines:
                    for line in lines:
                        name = line.strip()
                        chrom, coords = name.split(":")
                        start, end = [int(c) for c in coords.split("-")]
                        start += 1
                        start -= extend_up
                        end += extend_down
                        seq = self.get_seq(chrom, start, end)
                        yield Sequence(name, seq.seq)
    
                    lines = fin.readlines(BUFSIZE)
    
    
    
    def track2fasta(self, track, fastafile=None, stranded=False, extend_up=0, extend_down=0):
        track_type = get_track_type(track)
        if track_type == "interval":
            seqqer = self._region_to_seqs(track, extend_up=extend_up, extend_down=extend_down)
        else:
            seqqer = self._bed_to_seqs(
                    track, 
                    stranded=stranded, 
                    extend_up=extend_up, 
                    extend_down=extend_down)
        
        if fastafile:
            with open(fastafile, "w") as fout:
                for seq in seqqer:
                    fout.write("{}\n".format(seq.__repr__()))
        else:
            return [seq for seq in seqqer]

    def gap_sizes(self):
        """Return gap sizes per chromosome.
        
        Returns
        -------
        gap_sizes : dict
            a dictionary with chromosomes as key and the total number of
            Ns as values
        """ 
        if not self._gap_sizes:
            gap_file = self.props["gaps"]["gaps"]
            self._gap_sizes = {}
            with open(gap_file) as f:
                for line in f:
                    chrom, start, end = line.strip().split("\t")
                    start, end = int(start), int(end)
                    self._gap_sizes[chrom] = self._gap_sizes.get(chrom, 0) + end - start
        return self._gap_sizes
    
    def get_random_sequences(self, n=10, length=200, chroms=None, max_n=0.1):
        """Return random genomic sequences.

        Parameters
        ----------
        n : int , optional
            Number of sequences to return.

        length : int , optional
            Length of sequences to return.

        chroms : list , optional
            Return sequences only from these chromosomes.

        max_n : float , optional
            Maximum fraction of Ns.

        Returns
        -------
        coords : list
            List with [chrom, start, end] genomic coordinates.
        """ 
        retries = 100
        cutoff = length * max_n
        if not chroms:
            chroms = self.keys()
   
        try:
            gap_sizes = self.gap_sizes()
        except:
            gap_sizes = {}
        sizes = dict([(chrom, len(self[chrom]) - gap_sizes.get(chrom, 0)) for chrom in chroms])
   
        l = [(sizes[x], x) for x in chroms if 
                sizes[x] / len(self[x]) > 0.1 and sizes[x] > 10 * length]
        chroms = _weighted_selection(l, n)
        coords = []
        
        count = {}
        for chrom in chroms:
            if chrom in count:
                count[chrom] += 1
            else:
                count[chrom] = 1

        for chrom in chroms:
            for i in range(retries):
                start = int(random.random() * (sizes[chrom] - length))
                end = start + length
                count_n = self[chrom][start:end].seq.upper().count("N")
                if count_n <= cutoff:
                    break
            if count_n > cutoff:
                raise ValueError("Failed to find suitable non-N sequence for {}".format(chrom))
            
            coords.append([chrom, start, end])

        return coords

def manage_plugins(command, plugin_names=None):
    """Enable or disable plugins.
    """ 
    if plugin_names is None:
        plugin_names = []
    active_plugins = config.get("plugin", [])
    plugins = init_plugins()
    if command == "enable":
        for name in plugin_names:
            if name not in plugins:
                raise ValueError("Unknown plugin: {}".format(name))
            if name not in active_plugins:
                active_plugins.append(name)
    elif command == "disable":
        for name in plugin_names:
            if name in active_plugins:
                active_plugins.remove(name)
    elif command == "list":
        print("{:20}{}".format("plugin", "enabled"))
        for plugin in sorted(plugins):
            print("{:20}{}".format(plugin, {False:"", True:"*"}[plugin in active_plugins]))
    else:
        raise ValueError("Invalid plugin command")
    config["plugin"] = active_plugins
    config.save()

    if command in ["enable", "disable"]:
        print("Enabled plugins: {}".format(", ".join(sorted(active_plugins))))



#!/usr/bin/env python
import click
import os
import sys
import genomepy

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(genomepy.__about__.__version__)
def cli():
    """ Genomes for Python (and others)!
    
    Version: {}""".format(genomepy.__version__)
    pass

@click.command('search', short_help="search for genomes")
@click.argument("term")
@click.option("-p", "--provider", help="provider")
def search(term, provider=None):
    """Search for genomes that contain TERM in their name or description."""
    for row in genomepy.search(term, provider):
        print("\t".join([x.decode('utf-8', 'ignore') for x in row]))

@click.command('install', short_help="install genome")
@click.argument("name")
@click.argument("provider")
@click.option("-g", "--genome_dir", help="genome directory", default=None)
@click.option("-l", "--localname", help="custom name", default=None)
@click.option("-m", "--mask", help="mask (hard or soft)", default="soft")
@click.option("-r", "--regex", help="regex to filter sequences", default=None)
@click.option("--match/--no-match", help="set no-match to select sequences that *don't*  match regex", default=True)
@click.option("--annotation/--no-annotation", help="download annotation", default=False)
def install(name, provider, genome_dir, localname, mask, regex, match, annotation):
    """Install genome NAME from provider PROVIDER in directory GENOME_DIR."""
    genomepy.install_genome(
            name, provider, genome_dir=genome_dir, localname=localname, mask=mask, 
            regex=regex, invert_match=not(match), annotation=annotation)

@click.command('genomes', short_help="list available genomes")
@click.option("-p", "--provider", help="provider")
def genomes(provider=None):
    """List all available genomes."""
    for row in genomepy.list_available_genomes(provider):
        print("\t".join(row))

@click.command('providers', short_help="list available providers")
def providers():
    """List all available providers."""
    for p in genomepy.list_available_providers():
        print(p)

@click.command('plugin', short_help="manage plugins")
@click.argument("command")
@click.argument("name", nargs=-1)
def plugin(command, name):
    """Enable or disable plugins"""
    genomepy.functions.manage_plugins(command, name)

@click.command('config', short_help="manage configuration")
@click.argument("command")
def config(command):
    """Manage configuration"""
    genomepy.functions.manage_config(command)

cli.add_command(search)
cli.add_command(install)
cli.add_command(genomes)
cli.add_command(providers)
cli.add_command(plugin)
cli.add_command(config)

if __name__ == "__main__":
    cli()



from genomepy.plugin import Plugin

class SizesPlugin(Plugin):

    def after_genome_download(self, genome):
        props = self.get_properties(genome)
        fname = props["sizes"]
        with open(fname, "w") as f:
            for seqname in genome.keys():
                f.write("{}\t{}\n".format(seqname, len(genome[seqname])))

    def get_properties(self, genome):
        props = {
               "sizes":genome.filename + ".sizes",
               }
        return props



import re

from genomepy.plugin import Plugin

class GapsPlugin(Plugin):

    def after_genome_download(self, genome):
        props = self.get_properties(genome)
        fname = props["gaps"]
        
        with open(fname, "w") as bed:
            for chrom in genome.keys():
                for m in re.finditer(r'N+', genome[chrom][:].seq):
                    bed.write("{}\t{}\t{}\n".format(chrom, m.start(0), m.end(0)))
    def get_properties(self, genome):
        props = {
               "gaps": genome.filename.replace(".fa", ".gaps.bed"),
               }
        return props



"""Genome providers."""
import sys
import requests
import re
import os
import io
import time
import gzip
import xmltodict
import shutil
import subprocess as sp
import tarfile
import time
from tempfile import mkdtemp,NamedTemporaryFile
try:
        from urllib.request import urlopen, urlretrieve, urlcleanup
except:
        from urllib import urlopen, urlretrieve, urlcleanup

from bucketcache import Bucket
from pyfaidx import Fasta
from appdirs import user_cache_dir

from genomepy import exceptions
from genomepy.utils import filter_fasta, get_localname
from genomepy.__about__ import __version__

my_cache_dir = os.path.join(user_cache_dir("genomepy"), __version__)
# Create .cache dir if it does not exist
if not os.path.exists(my_cache_dir):
    os.makedirs(my_cache_dir)

cached = Bucket(my_cache_dir, days=7)

class ProviderBase(object):
    
    """Provider base class.

    Use to get a list of available providers:
    >>> ProviderBase.list_providers()
    ['UCSC', 'NCBI', 'Ensembl']

    Create a provider:
    >>> p = ProviderBase.create("UCSC")
    >>> for name, desc in p.search("hg38"):
    ...     print(desc)
    Human Dec. 2013 (GRCh38/hg38) Genome at UCSC
    """
    
    _providers = {}
    name = None

    @classmethod
    def create(cls, name):
        """Create a provider based on the provider name.

        Parameters
        ----------
        name : str
            Name of the provider (eg. UCSC, Ensembl, ...)
        
        Returns
        -------
        provider : Provider instance
            Provider instance.
        """
        try:
            return cls._providers[name]()
        except KeyError:
            raise Exception("Unknown provider")

    @classmethod
    def register_provider(cls, provider):
        """Register method to keep list of providers."""
        def decorator(subclass):
            """Register as decorator function."""
            cls._providers[provider] = subclass
            subclass.name = provider
            return subclass
        return decorator
    
    @classmethod
    def list_providers(self):
        """List available providers.""" 
        return self._providers.keys()

    def __hash__(self):
        return hash(str(self.__class__))
    
    def tar_to_bigfile(self, fname, outfile):
        """Convert tar of multiple FASTAs to one file."""
        fnames = []
        tmpdir = mkdtemp()
        
        # Extract files to temporary directory
        with tarfile.open(fname) as tar:
            tar.extractall(path=tmpdir)
        for root, _, files in os.walk(tmpdir):
            fnames += [os.path.join(root, fname) for fname in files]
        
        # Concatenate
        with open(outfile, "w") as out:
            for infile in fnames:
                for line in open(infile):
                    out.write(line)
                os.unlink(infile)
        
        # Remove temp dir
        shutil.rmtree(tmpdir)

    def download_genome(self, name, genome_dir, localname=None, mask="soft", regex=None, invert_match=False, version=None):
        """
        Download a (gzipped) genome file to a specific directory

        Parameters
        ----------
        name : str
            Genome / species name
        
        genome_dir : str
            Directory to install genome

        mask: str , optional
            Masking, soft, hard or none (all other strings)
        """
        genome_dir = os.path.expanduser(genome_dir)
        
        if not os.path.exists(genome_dir):
            os.makedirs(genome_dir)
        
        dbname, link = self.get_genome_download_link(name, mask=mask, version=version)
        myname = dbname 
        if localname:
            myname = localname
        
        myname = myname.replace(" ", "_")

        gzipped = False
        if link.endswith(".gz"):
            gzipped = True

        if not os.path.exists(os.path.join(genome_dir, myname)):
            os.makedirs(os.path.join(genome_dir, myname))
        urlcleanup()
        response = urlopen(link)
         
        sys.stderr.write("downloading from {}...\n".format(link))
        down_dir = genome_dir
        fname = os.path.join(genome_dir, myname, myname + ".fa")
        if regex:
            down_dir = mkdtemp()
            os.mkdir(os.path.join(down_dir, myname))
            fname = os.path.join(down_dir, myname, myname + ".fa") 
        with open(fname, "wb") as f_out:
            if gzipped:
                # Supports both Python 2.7 as well as 3
                with gzip.GzipFile(fileobj=io.BytesIO(response.read())) as f_in:
                    shutil.copyfileobj(f_in, f_out)
            else:
                f_out.write(response.read())
        sys.stderr.write("done...\n")
        
        if link.endswith("tar.gz"):
            self.tar_to_bigfile(fname, fname) 
        
        if hasattr(self, '_post_process_download'):
            self._post_process_download(name, down_dir, mask)
        
        if regex:
            infa = fname
            outfa = os.path.join(genome_dir, myname, myname + ".fa") 
            filter_fasta(
                infa,
                outfa,
                regex=regex,
                v=invert_match,
                force=True
                )

            not_included = [k for k in Fasta(infa).keys() if k not in Fasta(outfa).keys()]
            shutil.rmtree(down_dir)
            fname = outfa
        
        sys.stderr.write("name: {}\n".format(dbname))
        sys.stderr.write("local name: {}\n".format(myname))
        sys.stderr.write("fasta: {}\n".format(fname))

        # Create readme with information
        readme = os.path.join(genome_dir, myname, "README.txt")
        with open(readme, "w") as f:
            f.write("name: {}\n".format(myname))
            f.write("original name: {}\n".format(dbname))
            f.write("original filename: {}\n".format(os.path.split(link)[-1]))
            f.write("url: {}\n".format(link))
            f.write("mask: {}\n".format(mask))
            f.write("date: {}\n".format(time.strftime("%Y-%m-%d %H:%M:%S")))
            if regex:
                if invert_match:
                    f.write("regex: {} (inverted match)\n".format(regex))
                else:
                    f.write("regex: {}\n".format(regex))
                f.write("sequences that were excluded:\n")
                for seq in not_included:
                    f.write("\t{}\n".format(seq))
    
    def download_annotation(self, name, genome_dir, localname=None, version=None):
        """
        Download annotation file to to a specific directory

        Parameters
        ----------
        name : str
            Genome / species name
        
        genome_dir : str
            Directory to install annotation
        """
        raise NotImplementedError()

register_provider = ProviderBase.register_provider

@register_provider('Ensembl')
class EnsemblProvider(ProviderBase):
    
    """
    Ensembl genome provider.

    Will search both ensembl.org as well as ensemblgenomes.org.
    The bacteria division is not yet supported.
    """
    
    rest_url = "http://rest.ensembl.org/"
    
    def __init__(self):
        self.genomes = None
        self.version = None

    @cached(method=True)
    def request_json(self, ext):
        """Make a REST request and return as json."""
        if self.rest_url.endswith("/") and ext.startswith("/"):
            ext = ext[1:]
        
        r = requests.get(self.rest_url + ext, 
            headers={ "Content-Type" : "application/json"})

        if not r.ok:
            r.raise_for_status()
        
        return r.json()
    
    def safe(self, name):
        """Replace spaces with undescores.
        """
        return name.replace(" ", "_")

    def list_available_genomes(self, as_dict=False):
        """
        List all available genomes.
        
        Parameters
        ----------
        as_dict : bool, optional
            Return a dictionary of results.
        
        Yields
        ------
        genomes : dictionary or tuple
        """
        if not self.genomes:
            self.genomes = []
            divisions = self.request_json('info/divisions?')
            for division in divisions:
                genomes = self.request_json("info/genomes/division/{}?".format(division))
                self.genomes += genomes
       
        for genome in self.genomes:
            if as_dict:
                yield genome
            else:
                yield (self.safe(genome.get("assembly_name", "")), genome.get("name", ""))
   
    def search(self, term):
        """
        Search for a genome at Ensembl. 
        
        Both the name and description are used for the 
        search. Seacrch term is case-insensitive.
    
        Parameters
        ----------
        term : str
            Search term, case-insensitive.
    
        Yields
        ------
        tuple
            genome information (name/identifier and description)
        """
        term = term.lower()
        for genome in self.list_available_genomes(as_dict=True):
            if term in ",".join([str(v) for v in genome.values()]).lower():
                yield (self.safe(genome.get("assembly_name", "")), 
                        genome.get("name", ""))

    def _get_genome_info(self, name):
        """Get genome_info from json request."""
        try:
            assembly_acc = ""
            for genome in self.list_available_genomes(as_dict=True):
                if self.safe(genome.get("assembly_name", "")) == self.safe(name):
                    assembly_acc = genome.get("assembly_accession", "")
                    break
            if assembly_acc:
                ext  = "info/genomes/assembly/" + assembly_acc + "/?"
                genome_info = self.request_json(ext)
            else:
                raise exceptions.GenomeDownloadError(
                    "Could not download genome {} from Ensembl".format(name))
        except requests.exceptions.HTTPError as e:
            sys.stderr.write("Species not found: {}".format(e))
            raise exceptions.GenomeDownloadError(
                "Could not download genome {} from Ensembl".format(name))
        return genome_info

    def get_version(self, ftp_site):
        """Retrieve current version from Ensembl FTP.
        """
        response = urlopen(ftp_site + "/current_README")
        p = re.compile(r'Ensembl (Genomes|Release) (\d+)')
        m = p.search(response.read().decode())
        if m:
            version = m.group(2)
            sys.stderr.write("Using version {}\n".format(version))
            self.version = version
            return version

    def get_genome_download_link(self, name, version=None, mask="soft"):
        """
        Return Ensembl ftp link to toplevel genome sequence

        Parameters
        ----------
        name : str 
            Genome name. Current implementation will fail if exact
            name is not found.
        
        Returns
        ------
        tuple (name, link) where name is the Ensembl dbname identifier
        and link is a str with the ftp download link.
        """
        genome_info = self._get_genome_info(name)
        
        # parse the division
        division = genome_info["division"].lower().replace("ensembl","")
        if division == "bacteria":
            raise NotImplementedError("bacteria from ensembl not yet supported")
        
        ftp_site = "ftp://ftp.ensemblgenomes.org/pub"
        if division == "vertebrates":
            ftp_site = "https://ftp.ensembl.org/pub"

        version = self.version
        if not version:
            version = self.get_version(ftp_site)

        if division != 'vertebrates':
            base_url = "/{}/release-{}/fasta/{}/dna/"
            ftp_dir = base_url.format(division, version, genome_info["url_name"].lower())
            url = "{}/{}".format(ftp_site, ftp_dir)
        else:
            base_url = "/release-{}/fasta/{}/dna/"
            ftp_dir = base_url.format(version, genome_info["url_name"].lower())
            url = "{}/{}".format(ftp_site, ftp_dir)
      
        pattern = "dna.toplevel"
        if mask == "soft":
            pattern = "dna_sm.toplevel"
        elif mask == "hard":
            pattern = "dna_rm.toplevel"
        
        asm_url = "{}/{}.{}.{}.fa.gz".format(
            url,
            genome_info['url_name'].capitalize(),
            re.sub(r'\.p\d+$', '', self.safe(genome_info["assembly_name"])),
            pattern)
        
        return self.safe(genome_info["assembly_name"]), asm_url

    def download_annotation(self, name, genome_dir, localname=None, version=None):
        """
        Download gene annotation from Ensembl based on genome name.
    
        Parameters
        ----------
        name : str
            Ensembl genome name.
        genome_dir : str
            Genome directory.
        version : str , optional
            Ensembl version. By default the latest version is used.
        """
        localname = get_localname(name, localname)
        genome_info = self._get_genome_info(name)

        # parse the division
        division = genome_info["division"].lower().replace("ensembl","")
        if division == "bacteria":
            raise NotImplementedError("bacteria from ensembl not yet supported")
                
        # Get the base link depending on division
        ftp_site = "https://ftp.ensemblgenomes.org/pub/{}".format(division)
        if division == 'vertebrates':
            ftp_site = "https://ftp.ensembl.org/pub"

        if not version:
            version = self.get_version(ftp_site)

        # Get the GTF URL
        base_url = ftp_site + "/release-{}/gtf/{}/{}.{}.{}.gtf.gz"
        safe_name = name.replace(" ", "_")
        safe_name = re.sub('\.p\d+$', '', safe_name)

        ftp_link = base_url.format(version, genome_info["url_name"].lower(), genome_info["url_name"].capitalize(), safe_name, version)

        out_dir = os.path.join(genome_dir, localname)
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)
        # Download the file
        try:
            response = urlopen(ftp_link)
            gtf_file = out_dir + "/"+ localname + ".annotation.gtf.gz"
            with open(gtf_file, "wb") as f:
                f.write(response.read())
        
            bed_file = gtf_file.replace("gtf.gz", "bed")
            cmd = "gtfToGenePred {0} /dev/stdout | genePredToBed /dev/stdin {1} && gzip {1}"
            ret = sp.check_call(cmd.format(gtf_file, bed_file), shell=True)
            readme = os.path.join(genome_dir, localname, "README.txt")
            with open(readme, "a") as f:
                f.write("annotation url: {}\n".format(ftp_link))
        except Exception:
            sys.stderr.write("\nCould not download {}\n".format(ftp_link))
            raise

        return out_dir

@register_provider('UCSC')
class UcscProvider(ProviderBase):
     
    """
    UCSC genome provider.
    
    The UCSC DAS server is used to search and list genomes. 
    """
    
    ucsc_url = "http://hgdownload.soe.ucsc.edu/goldenPath/{0}/bigZips/chromFa.tar.gz"
    ucsc_url_masked = "http://hgdownload.soe.ucsc.edu/goldenPath/{0}/bigZips/chromFaMasked.tar.gz"
    alt_ucsc_url = "http://hgdownload.soe.ucsc.edu/goldenPath/{0}/bigZips/{0}.fa.gz"
    alt_ucsc_url_masked = "http://hgdownload.soe.ucsc.edu/goldenPath/{0}/bigZips/{0}.fa.masked.gz"
    das_url = "http://genome.ucsc.edu/cgi-bin/das/dsn"
    
    def __init__(self):
        self.genomes = []
   
    def list_available_genomes(self):
        """
        List all available genomes.
        
        Returns
        -------
        genomes : list
        """
        self.genomes = self._get_genomes()
        
        return self.genomes
    
    @cached(method=True) 
    def _get_genomes(self):
        response = urlopen(self.das_url)
        d = xmltodict.parse(response.read())
        genomes = []
        for genome in d['DASDSN']['DSN']:
                genomes.append([
                    genome['SOURCE']['@id'], genome['DESCRIPTION']
                    ])
        return genomes


    def search(self, term):
        """
        Search for a genome at UCSC. 
        
        Both the name and description are used for the 
        search. Search term is case-insensitive.
    
        Parameters
        ----------
        term : str
            Search term, case-insensitive.
    
        Yields
        ------
        tuple
            genome information (name/identifier and description)
        """
        term = term.lower()
        for name,description in self.list_available_genomes():
            if term in name.lower() or term in description.lower():
                yield name,description
    
    def get_genome_download_link(self, name, mask="soft", version=None):
        """
        Return UCSC http link to genome sequence

        Parameters
        ----------
        name : str 
            Genome build name. Current implementation will fail if exact
            name is not found.
        
        Returns
        ------
        tuple (name, link) where name is the genome build identifier
        and link is a str with the http download link.
        """
        urls = [self.ucsc_url, self.alt_ucsc_url]
        if mask == "hard":
            urls = [self.ucsc_url_masked, self.alt_ucsc_url_masked]

        for genome_url in urls:
            remote = genome_url.format(name)
            ret = requests.head(remote)
            
            if ret.status_code == 200:
                return name, remote

        raise exceptions.GenomeDownloadError(
                "Could not download genome {} from UCSC".format(name))


    def download_annotation(self, name, genome_dir, localname=None, version=None):
        """
        Download gene annotation from UCSC based on genomebuild.
    
        Will check UCSC, Ensembl and RefSeq annotation.
    
        Parameters
        ----------
        genomebuild : str
            UCSC genome name.
        genome_dir : str
            Genome directory.
        """
        localname = get_localname(name, localname)

        UCSC_GENE_URL = "http://hgdownload.cse.ucsc.edu/goldenPath/{}/database/"
        ANNOS = ["knownGene.txt.gz", "ensGene.txt.gz", "refGene.txt.gz"]
        pred = "genePredToBed"
    
        tmp = NamedTemporaryFile(delete=False, suffix=".gz")
    
        anno = []
        f = urlopen(UCSC_GENE_URL.format(name))
        p = re.compile(r'\w+.Gene.txt.gz')
        for line in f.readlines():
            m = p.search(line.decode())
            if m:
                anno.append(m.group(0))
        sys.stderr.write("Retrieving gene annotation for {}\n".format(name))
        url = ""
        for a in ANNOS:
            if a in anno:
                url = UCSC_GENE_URL.format(name) + a
                break
        if url:
            sys.stderr.write("Using {}\n".format(url))
            urlretrieve(
                    url,
                    tmp.name
                    )
             
            with gzip.open(tmp.name) as f:
                cols = f.readline().decode(errors='ignore').split("\t")
    
            start_col = 1
            for i,col in enumerate(cols):
                if col == "+" or col == "-":
                    start_col = i - 1
                    break
            end_col = start_col + 10
           
            localname = localname.replace(" ", "_")
            path = os.path.join(genome_dir, localname)
            if not os.path.exists(path):
                os.mkdir(path)
            bed_file = os.path.join(genome_dir, localname, localname + ".annotation.bed")
            cmd = "zcat {} | cut -f{}-{} | {} /dev/stdin {} && gzip {}"
            sp.call(cmd.format(
                tmp.name, start_col, end_col, pred, bed_file, bed_file), 
                shell=True)
            
            gtf_file = bed_file.replace(".bed", ".gtf")
            cmd = "bedToGenePred {0}.gz /dev/stdout | genePredToGtf file /dev/stdin /dev/stdout -utr -honorCdsStat | sed 's/.dev.stdin/UCSC/' > {1} && gzip {1}"
            ret = sp.check_call(cmd.format(bed_file, gtf_file), shell=True)
        
            readme = os.path.join(genome_dir, localname, "README.txt")
            with open(readme, "a") as f:
                f.write("annotation url: {}\n".format(url))
        else:
            sys.stderr.write("No annotation found!")

@register_provider('NCBI')
class NCBIProvider(ProviderBase):
        
    """
    NCBI genome provider.
    
    Useas the assembly reports page to search and list genomes. 
    """
    
    assembly_url = "https://ftp.ncbi.nlm.nih.gov/genomes/ASSEMBLY_REPORTS/"

    def __init__(self):
        self.genomes = None

    @cached(method=True)
    def _get_genomes(self):
        """Parse genomes from assembly summary txt files."""
        genomes = []
        
        names = [
                "assembly_summary_refseq.txt", 
                "assembly_summary_genbank.txt", 
                "assembly_summary_refseq_historical.txt",
                ]
        
        sys.stderr.write("Downloading assembly summaries from NCBI, this will take a while...\n")
        seen = {}
        for fname in names:
            urlcleanup() 
            response = urlopen(self.assembly_url + "/" + fname)
            lines = response.read().decode('utf-8').splitlines()
            header = lines[1].strip("# ").split("\t")
            for line in lines[2:]:
                vals = line.strip("# ").split("\t")
                # Don't repeat samples with the same BioSample ID
                if vals[2] not in seen:   # BioSample ID
                    genomes.append(dict(zip(header, vals)))
                    seen[vals[2]] = 1
        
        return genomes

    def list_available_genomes(self, as_dict=False):
        """
        List all available genomes.
        
        Parameters
        ----------
        as_dict : bool, optional
            Return a dictionary of results.
        
        Yields
        ------
        genomes : dictionary or tuple
        """
        if not self.genomes:
            self.genomes = self._get_genomes()

        for genome in self.genomes:
            if as_dict:
                yield genome
            else:
                yield (
                        genome.get("asm_name", ""), 
                        "; ".join((
                            genome.get("organism_name", ""),
                            genome.get("submitter", ""),
                        ))
                    )
   
    def search(self, term):
        """
        Search for term in genome names and descriptions. 

        The search is case-insensitive.

        Parameters
        ----------
        term : str
            search term
        
        Yields
        ------
        tuples with two items, name and description
        """
        term = term.lower()

        for genome in self.list_available_genomes(as_dict=True):
            term_str = ";".join([repr(x) for x in genome.values()])

            if term in term_str.lower():
                yield (
                        genome.get("asm_name", ""), 
                        "; ".join((
                            genome.get("organism_name", ""),
                            genome.get("submitter", ""),
                        ))
                    )

    def get_genome_download_link(self, name, mask="soft", version=None):
        """
        Return NCBI ftp link to top-level genome sequence

        Parameters
        ----------
        name : str 
            Genome name. Current implementation will fail if exact
            name is not found.
        
        Returns
        ------
        tuple (name, link) where name is the NCBI asm_name identifier
        and link is a str with the ftp download link.
        """
        if mask == "hard":
            sys.stderr.write("ignoring mask parameter for NCBI at download.\n")

        if not self.genomes:
            self.genomes = self._get_genomes()
        
        for genome in self.genomes:
            if genome["asm_name"] == name:
                #ftp_path': 'ftp://ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/004/195/GCF_000004195.3_Xenopus_tropicalis_v9.1'
                url = genome["ftp_path"]
                url = url.replace("ftp://", "https://")
                url += "/" + url.split("/")[-1] + "_genomic.fna.gz"
                return name, url
        raise exceptions.GenomeDownloadError("Could not download genome from NCBI")
    
    def _post_process_download(self, name, genome_dir, mask="soft"):
        """
        Replace accessions with sequence names in fasta file.
        
        Parameters
        ----------
        name : str
            NCBI genome name

        genome_dir : str
            Genome directory
        """
        # Get the FTP url for this specific genome and download
        # the assembly report
        for genome in self.genomes:
            if genome["asm_name"] == name:
                url = genome["ftp_path"]
                url += "/" + url.split("/")[-1] + "_assembly_report.txt"
                url = url.replace("ftp://", "https://")
                break
   
        # Create mapping of accessions to names
        tr = {}
        urlcleanup()
        response = urlopen(url)
        for line in response.read().decode('utf-8').splitlines():
            if line.startswith("#"):
                continue
            vals = line.strip().split("\t")
            tr[vals[6]] = vals[0]
    
        name = name.replace(" ", "_")
        # Check of the original genome fasta exists
        fa = os.path.join(genome_dir, name, "{}.fa".format(name))
        if not os.path.exists(fa):
            raise Exception("Genome fasta file not found, {}".format(fa))
        
        # Use a tmp file and replace the names
        new_fa = os.path.join(
                genome_dir, name,
                ".process.{}.fa".format(name)
                )
        if mask == "hard":
            sys.stderr.write("masking lower-case.\n")

        with open(fa) as old:
            with open(new_fa, "w") as new:
                for line in old:
                    if line.startswith(">"):
                        desc = line.strip()[1:]
                        name = desc.split(" ")[0]
                        new.write(">{} {}\n".format(
                            tr.get(name, name),
                            desc
                            ))
                    elif mask == "hard":
                        new.write(re.sub('[actg]', 'N', line))
                    else:
                        new.write(line)
        
        # Rename tmp file to real genome file
        shutil.move(new_fa, fa)

    def download_annotation(self, name, genome_dir, localname=None, version=None):
        """
        Download annotation file to to a specific directory

        Parameters
        ----------
        name : str
            Genome / species name
        
        genome_dir : str
            Directory to install annotation
        """
        sys.stderr.write("Downloading annotation...\n")
        if not self.genomes:
            self.genomes = self._get_genomes()
        
        for genome in self.genomes:
            if genome["asm_name"] == name:
                #ftp_path': 'ftp://ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/004/195/GCF_000004195.3_Xenopus_tropicalis_v9.1'
                url = genome["ftp_path"]
                url += "/" + url.split("/")[-1] + "_genomic.gff.gz"
  
        localname = get_localname(name, localname)
        out_dir = os.path.join(genome_dir, localname)
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)
        
        # Download the file
        try:
            response = urlopen(url)
            gff_file = out_dir + "/"+ localname + ".annotation.gff.gz"
            with open(gff_file, "wb") as f:
                f.write(response.read())
        except Exception:
            sys.stderr.write("WARNING: Could not download annotation from NCBI, skipping.\n")
            sys.stderr.write("URL: {}\n".format(url))

            sys.stderr.write("If you think the annotation should be there, please file a bug report at:\n")
            sys.stderr.write("https://github.com/simonvh/genomepy/issues\n")
            return
        
        cmd = "gff3ToGenePred {0} /dev/stdout | wc -l"
        out = sp.check_output(cmd.format(gff_file), shell=True)
        if out.strip() == b"0":
            sys.stderr.write("WARNING: annotation from NCBI contains no genes, skipping.\n")
        else:
            # Convert to BED file
            bed_file = gff_file.replace("gff.gz", "bed")
            cmd = "gff3ToGenePred -rnaNameAttr=gene {0} /dev/stdout | genePredToBed /dev/stdin {1} && gzip {1}"
            ret = sp.check_call(cmd.format(gff_file, bed_file), shell=True) 

            # Convert to GTF file
            gtf_file = gff_file.replace("gff.gz", "gtf")
            cmd = "gff3ToGenePred -geneNameAttr=gene {0} /dev/stdout | genePredToGtf file /dev/stdin {1} && gzip {1}"
            ret = sp.check_call(cmd.format(gff_file, gtf_file), shell=True)
        
        readme = os.path.join(genome_dir, name, "README.txt")
        with open(readme, "a") as f:
            f.write("annotation url: {}\n".format(url))

        return out_dir



